import logging

import numpy
import numpy.random
import pytest

from grpc4bmi.bmi_grpc_legacy_server import BmiLegacyServer02
from test.flatbmiheat import FlatLegacyBmiHeat

"""
Unit tests for the BMI server class. Every test performs cross-checking with a local instance of the BMI heat toy model.
"""

logging.basicConfig(level=logging.DEBUG)


class RequestStub(object):
    def __init__(self):
        self.config_file = ""

    def HasField(self, name):
        return hasattr(self, name)


class value_wrapper(object):
    def __init__(self, vals):
        self.values = vals.flatten()


def make_string(obj):
    return '' if obj is None else str(obj)


def make_list(obj):
    if obj is list:
        return obj
    if obj is None:
        return []
    return [obj]


def make_bmi_classes(init=False):
    server, local = BmiLegacyServer02(FlatLegacyBmiHeat()), FlatLegacyBmiHeat()
    if init:
        req = RequestStub()
        numpy.random.seed(0)
        server.initialize(req, None)
        numpy.random.seed(0)
        local.initialize(None)
    return server, local


def test_server_start():
    server, local = make_bmi_classes()
    assert server is not None
    del server


def test_component_name():
    server, local = make_bmi_classes()
    assert server.getComponentName(None, None).name == local.get_component_name()
    del server


def test_varnames():
    server, local = make_bmi_classes()
    assert server.getInputVarNames(None, None).names == list(local.get_input_var_names())
    assert server.getOutputVarNames(None, None).names == list(local.get_output_var_names())
    del server


def test_initialize():
    server, local = make_bmi_classes(True)
    assert server is not None
    server.finalize(None, None)
    del server


def test_update():
    server, local = make_bmi_classes(True)
    server.update(None, None)
    assert server is not None
    server.finalize(None, None)
    del server


def test_get_time_unit():
    server, local = make_bmi_classes()
    assert server.getTimeUnits(None, None).units == make_string(local.get_time_units())
    server.finalize(None, None)
    del server


def test_get_time_step():
    server, local = make_bmi_classes(True)
    assert server.getTimeStep(None, None).interval == local.get_time_step()
    server.finalize(None, None)
    del server


def test_get_current_time():
    server, local = make_bmi_classes(True)
    assert server.getCurrentTime(None, None).time == local.get_current_time()
    server.finalize(None, None)
    del server


def test_get_updated_time():
    server, local = make_bmi_classes(True)
    server.update(None, None)
    assert server.getCurrentTime(None, None).time != local.get_current_time()
    local.update()
    assert server.getCurrentTime(None, None).time == local.get_current_time()
    server.finalize(None, None)
    del server


def test_get_start_end_time():
    server, local = make_bmi_classes(True)
    assert server.getStartTime(None, None).time == local.get_start_time()
    assert server.getEndTime(None, None).time == local.get_end_time()
    server.finalize(None, None)
    del server


def test_get_var_grid():
    server, local = make_bmi_classes(True)
    request = RequestStub()
    varname = local.get_output_var_names()[0]
    setattr(request, "name", varname)
    assert server.getVarGrid(request, None).grid_id == local.get_var_grid(varname)
    del server


def test_get_var_type():
    server, local = make_bmi_classes(True)
    request = RequestStub()
    varname = local.get_output_var_names()[0]
    setattr(request, "name", varname)
    assert server.getVarType(request, None).type == local.get_var_type(varname)
    del server


def test_get_var_units():
    server, local = make_bmi_classes(True)
    request = RequestStub()
    varname = local.get_output_var_names()[0]
    setattr(request, "name", varname)
    assert server.getVarUnits(request, None).units == local.get_var_units(varname)
    del server


def test_get_var_nbytes():
    server, local = make_bmi_classes(True)
    request = RequestStub()
    varname = local.get_output_var_names()[0]
    setattr(request, "name", varname)
    assert server.getVarNBytes(request, None).nbytes == local.get_var_nbytes(varname)
    del server


def test_get_var_values():
    server, local = make_bmi_classes(True)
    request = RequestStub()
    varname = local.get_output_var_names()[0]
    setattr(request, "name", varname)
    values = local.get_value(varname)
    numpy.testing.assert_allclose(numpy.reshape(server.getValue(request, None).values_double.values, values.shape), values)


def test_get_var_ptr():
    server, local = make_bmi_classes(True)
    request = RequestStub()
    varname = local.get_output_var_names()[0]
    setattr(request, "name", varname)
    with pytest.raises(NotImplementedError):
        server.getValuePtr(request, None)


def test_get_vals_indices():
    server, local = make_bmi_classes(True)
    request = RequestStub()
    varname = local.get_output_var_names()[0]
    indices = numpy.array([29, 8, 19, 81])
    setattr(request, "name", varname)
    setattr(request, "indices", indices.flatten())
    setattr(request, "index_size", 1)
    values = local.get_value_at_indices(varname, indices)
    numpy.testing.assert_allclose(server.getValueAtIndices(request, None).values_double.values, values.flatten())


def test_get_vals_indices_2d():
    server, local = make_bmi_classes(True)
    request = RequestStub()
    varname = local.get_output_var_names()[0]
    indices = numpy.array([[0, 1], [1, 0], [2, 2]])
    setattr(request, "name", varname)
    setattr(request, "indices", indices.flatten())
    setattr(request, "index_size", 2)
    values = local.get_value_at_indices(varname, indices)
    numpy.testing.assert_allclose(server.getValueAtIndices(request, None).values_double.values, values.flatten())


def test_set_var_values():
    server, local = make_bmi_classes(True)
    request = RequestStub()
    varname = local.get_output_var_names()[0]
    values = 0.123 * local.get_value(varname)
    setattr(request, "name", varname)
    setattr(request, "values_double", value_wrapper(values))
    setattr(request, "shape", values.shape)
    server.setValue(request, None)
    delattr(request, "values_double")
    delattr(request, "shape")
    response = server.getValue(request, None)
    numpy.testing.assert_allclose(values, response.values_double.values)


def test_set_values_indices():
    server, local = make_bmi_classes(True)
    request = RequestStub()
    varname = local.get_output_var_names()[0]
    indices = numpy.array([1, 11, 21])
    values = numpy.array([0.123, 4.567, 8.901])
    setattr(request, "name", varname)
    setattr(request, "values_double", value_wrapper(values))
    setattr(request, "indices", indices.flatten())
    setattr(request, "index_size", 1)
    server.setValueAtIndices(request, None)
    delattr(request, "values_double")
    response = server.getValueAtIndices(request, None)
    values_copy = numpy.array(response.values_double.values)
    numpy.testing.assert_allclose(values, values_copy)


def test_get_grid_size():
    server, local = make_bmi_classes(True)
    request = RequestStub()
    varname = local.get_output_var_names()[0]
    grid_id = local.get_var_grid(varname)
    setattr(request, "grid_id", grid_id)
    assert server.getGridSize(request, None).size == local.get_grid_size(grid_id)


def test_get_grid_rank():
    server, local = make_bmi_classes(True)
    request = RequestStub()
    varname = local.get_output_var_names()[0]
    grid_id = local.get_var_grid(varname)
    setattr(request, "grid_id", grid_id)
    assert server.getGridRank(request, None).rank == local.get_grid_rank(grid_id)


def test_get_grid_type():
    server, local = make_bmi_classes(True)
    request = RequestStub()
    varname = local.get_output_var_names()[0]
    grid_id = local.get_var_grid(varname)
    setattr(request, "grid_id", grid_id)
    assert server.getGridType(request, None).type == local.get_grid_type(grid_id)


def test_get_grid_shape():
    server, local = make_bmi_classes(True)
    request = RequestStub()
    varname = local.get_output_var_names()[0]
    grid_id = local.get_var_grid(varname)
    setattr(request, "grid_id", grid_id)
    assert tuple(server.getGridShape(request, None).shape) == local.get_grid_shape(grid_id)


def test_get_grid_spacing():
    server, local = make_bmi_classes(True)
    request = RequestStub()
    varname = local.get_output_var_names()[0]
    grid_id = local.get_var_grid(varname)
    setattr(request, "grid_id", grid_id)
    assert tuple(server.getGridSpacing(request, None).spacing) == local.get_grid_spacing(grid_id)


def test_get_grid_origin():
    server, local = make_bmi_classes(True)
    request = RequestStub()
    varname = local.get_output_var_names()[0]
    grid_id = local.get_var_grid(varname)
    setattr(request, "grid_id", grid_id)
    assert tuple(server.getGridOrigin(request, None).origin) == local.get_grid_origin(grid_id)



import logging

import numpy
from bmipy import Bmi
from grpc_status import rpc_status
from google.protobuf import any_pb2
from google.rpc import code_pb2, status_pb2, error_details_pb2
import traceback

from grpc4bmi.reserve import reserve_values, reserve_grid_shape, reserve_grid_nodes, reserve_grid_padding, \
    reserve_values_at_indices
from . import bmi_pb2, bmi_pb2_grpc

log = logging.getLogger(__name__)


class BmiServer(bmi_pb2_grpc.BmiServiceServicer):
    """
    BMI Server class, wrapping an existing python implementation and exposing it via GRPC across the memory space (to
    listening client processes). The class takes a package, module and class name and instantiates the BMI
    implementation by assuming a default constructor with no arguments.

    Args:
        model: Bmi model object which must be wrapped by grpc
        debug: If true then returns stacktrace in an error response.
                The stacktrace is returned in the trailing metadata as a DebugInfo (https://github.com/googleapis/googleapis/blob/07244bb797ddd6e0c1c15b02b4467a9a5729299f/google/rpc/error_details.proto#L46-L52) message.
    """

    def __init__(self, model, debug=False):
        # type: (BmiServer, Bmi, bool) -> None
        super(bmi_pb2_grpc.BmiServiceServicer, self).__init__()
        self.bmi_model_ = model
        self.debug = debug

    def exception_handler(self, exc, context):
        log.exception(exc)
        detail = any_pb2.Any()
        if self.debug:
            detail.Pack(
                error_details_pb2.DebugInfo(
                    stack_entries=traceback.format_stack(),
                    detail=repr(exc)
                )
            )
        status = status_pb2.Status(
            code=code_pb2.INTERNAL,
            message=str(exc),
            details=[detail]
        )
        context.abort_with_status(rpc_status.to_status(status))

    def initialize(self, request, context):
        ifile = str(request.config_file)
        if not ifile:
            ifile = None
        try:
            self.bmi_model_.initialize(ifile)
            return bmi_pb2.Empty()
        except Exception as e:
            self.exception_handler(e, context)

    def update(self, request, context):
        try:
            self.bmi_model_.update()
            return bmi_pb2.Empty()
        except Exception as e:
            self.exception_handler(e, context)

    def updateUntil(self, request, context):
        try:
            self.bmi_model_.update_until(request.time)
            return bmi_pb2.Empty()
        except Exception as e:
            self.exception_handler(e, context)

    def finalize(self, request, context):
        try:
            self.bmi_model_.finalize()
            return bmi_pb2.Empty()
        except Exception as e:
            self.exception_handler(e, context)

    def getComponentName(self, request, context):
        try:
            return bmi_pb2.GetComponentNameResponse(name=self.bmi_model_.get_component_name())
        except Exception as e:
            self.exception_handler(e, context)

    def getInputItemCount(self, request, context):
        try:
            return bmi_pb2.GetCountResponse(count=self.bmi_model_.get_input_item_count())
        except Exception as e:
            self.exception_handler(e, context)

    def getOutputItemCount(self, request, context):
        try:
            return bmi_pb2.GetCountResponse(count=self.bmi_model_.get_output_item_count())
        except Exception as e:
            self.exception_handler(e, context)

    def getInputVarNames(self, request, context):
        try:
            return bmi_pb2.GetVarNamesResponse(names=self.bmi_model_.get_input_var_names())
        except Exception as e:
            self.exception_handler(e, context)

    def getOutputVarNames(self, request, context):
        try:
            return bmi_pb2.GetVarNamesResponse(names=self.bmi_model_.get_output_var_names())
        except Exception as e:
            self.exception_handler(e, context)

    def getTimeUnits(self, request, context):
        try:
            return bmi_pb2.GetTimeUnitsResponse(units=self.bmi_model_.get_time_units())
        except Exception as e:
            self.exception_handler(e, context)

    def getTimeStep(self, request, context):
        try:
            return bmi_pb2.GetTimeStepResponse(interval=self.bmi_model_.get_time_step())
        except Exception as e:
            self.exception_handler(e, context)

    def getCurrentTime(self, request, context):
        try:
            return bmi_pb2.GetTimeResponse(time=self.bmi_model_.get_current_time())
        except Exception as e:
            self.exception_handler(e, context)

    def getStartTime(self, request, context):
        try:
            return bmi_pb2.GetTimeResponse(time=self.bmi_model_.get_start_time())
        except Exception as e:
            self.exception_handler(e, context)

    def getEndTime(self, request, context):
        try:
            return bmi_pb2.GetTimeResponse(time=self.bmi_model_.get_end_time())
        except Exception as e:
            self.exception_handler(e, context)

    def getVarGrid(self, request, context):
        try:
            return bmi_pb2.GetVarGridResponse(grid_id=self.bmi_model_.get_var_grid(request.name))
        except Exception as e:
            self.exception_handler(e, context)

    def getVarType(self, request, context):
        try:
            return bmi_pb2.GetVarTypeResponse(type=self.bmi_model_.get_var_type(request.name))
        except Exception as e:
            self.exception_handler(e, context)

    def getVarItemSize(self, request, context):
        try:
            return bmi_pb2.GetVarItemSizeResponse(size=self.bmi_model_.get_var_itemsize(request.name))
        except Exception as e:
            self.exception_handler(e, context)

    def getVarUnits(self, request, context):
        try:
            return bmi_pb2.GetVarUnitsResponse(units=self.bmi_model_.get_var_units(request.name))
        except Exception as e:
            self.exception_handler(e, context)

    def getVarNBytes(self, request, context):
        try:
            return bmi_pb2.GetVarNBytesResponse(nbytes=self.bmi_model_.get_var_nbytes(request.name))
        except Exception as e:
            self.exception_handler(e, context)

    def getVarLocation(self, request, context):
        location_name = self.bmi_model_.get_var_location(request.name)
        location = bmi_pb2.GetVarLocationResponse.Location.Value(location_name.upper())
        return bmi_pb2.GetVarLocationResponse(location=location)

    def getValue(self, request, context):
        try:
            values = reserve_values(self.bmi_model_, request.name)
            values = self.bmi_model_.get_value(request.name, values)
            if values.dtype in (numpy.int64, numpy.int32, numpy.int16):
                return bmi_pb2.GetValueResponse(values_int=bmi_pb2.IntArrayMessage(values=values.flatten()))
            if values.dtype in (numpy.float32, numpy.float16):
                return bmi_pb2.GetValueResponse(values_float=bmi_pb2.FloatArrayMessage(values=values.flatten()))
            if values.dtype == numpy.float64:
                return bmi_pb2.GetValueResponse(values_double=bmi_pb2.DoubleArrayMessage(values=values.flatten()))
            raise NotImplementedError("Arrays with type %s cannot be transmitted through this GRPC channel" % values.dtype)
        except Exception as e:
            self.exception_handler(e, context)

    def getValuePtr(self, request, context):
        raise NotImplementedError("Array references cannot be transmitted through this GRPC channel")

    def getValueAtIndices(self, request, context):
        try:
            indices = numpy.array(request.indices)
            values = reserve_values_at_indices(self.bmi_model_, request.name, indices)
            values = self.bmi_model_.get_value_at_indices(request.name, values, indices)
            if values.dtype in (numpy.int64, numpy.int32, numpy.int16):
                return bmi_pb2.GetValueAtIndicesResponse(values_int=bmi_pb2.IntArrayMessage(values=values.flatten()))
            if values.dtype in (numpy.float32, numpy.float16):
                return bmi_pb2.GetValueAtIndicesResponse(values_float=bmi_pb2.FloatArrayMessage(values=values.flatten()))
            if values.dtype == numpy.float64:
                return bmi_pb2.GetValueAtIndicesResponse(values_double=bmi_pb2.DoubleArrayMessage(values=values.flatten()))
            raise NotImplementedError("Arrays with type %s cannot be transmitted through this GRPC channel" % values.dtype)
        except Exception as e:
            self.exception_handler(e, context)

    def setValue(self, request, context):
        try:
            if request.HasField("values_int"):
                array = numpy.array(request.values_int.values, dtype=numpy.int64)
                self.bmi_model_.set_value(request.name, array)
            if request.HasField("values_float"):
                array = numpy.array(request.values_float.values, dtype=numpy.float32)
                self.bmi_model_.set_value(request.name, array)
            if request.HasField("values_double"):
                array = numpy.array(request.values_double.values, dtype=numpy.float64)
                self.bmi_model_.set_value(request.name, array)
            return bmi_pb2.Empty()
        except Exception as e:
            self.exception_handler(e, context)

    def setValueAtIndices(self, request, context):
        try:
            index_array = numpy.array(request.indices)
            if request.HasField("values_int"):
                array = numpy.array(request.values_int.values, dtype=numpy.int64)
                self.bmi_model_.set_value_at_indices(request.name, index_array, array)
            if request.HasField("values_float"):
                array = numpy.array(request.values_float.values, dtype=numpy.float32)
                self.bmi_model_.set_value_at_indices(request.name, index_array, array)
            if request.HasField("values_double"):
                array = numpy.array(request.values_double.values, dtype=numpy.float64)
                self.bmi_model_.set_value_at_indices(request.name, index_array, array)
            return bmi_pb2.Empty()
        except Exception as e:
            self.exception_handler(e, context)

    def getGridSize(self, request, context):
        try:
            return bmi_pb2.GetGridSizeResponse(size=self.bmi_model_.get_grid_size(request.grid_id))
        except Exception as e:
            self.exception_handler(e, context)

    def getGridRank(self, request, context):
        try:
            return bmi_pb2.GetGridRankResponse(rank=self.bmi_model_.get_grid_rank(request.grid_id))
        except Exception as e:
            self.exception_handler(e, context)

    def getGridType(self, request, context):
        try:
            return bmi_pb2.GetGridTypeResponse(type=self.bmi_model_.get_grid_type(request.grid_id))
        except Exception as e:
            self.exception_handler(e, context)

    def getGridShape(self, request, context):
        try:
            values = reserve_grid_shape(self.bmi_model_, request.grid_id)
            return bmi_pb2.GetGridShapeResponse(shape=self.bmi_model_.get_grid_shape(request.grid_id, values))
        except Exception as e:
            self.exception_handler(e, context)

    def getGridSpacing(self, request, context):
        try:
            values = reserve_grid_padding(self.bmi_model_, request.grid_id)
            return bmi_pb2.GetGridSpacingResponse(spacing=self.bmi_model_.get_grid_spacing(request.grid_id, values))
        except Exception as e:
            self.exception_handler(e, context)

    def getGridOrigin(self, request, context):
        try:
            values = reserve_grid_padding(self.bmi_model_, request.grid_id)
            return bmi_pb2.GetGridOriginResponse(origin=self.bmi_model_.get_grid_origin(request.grid_id, values))
        except Exception as e:
            self.exception_handler(e, context)

    def getGridX(self, request, context):
        try:
            values = reserve_grid_nodes(self.bmi_model_, request.grid_id, 0)
            return bmi_pb2.GetGridPointsResponse(coordinates=self.bmi_model_.get_grid_x(request.grid_id, values))
        except Exception as e:
            self.exception_handler(e, context)

    def getGridY(self, request, context):
        try:
            values = reserve_grid_nodes(self.bmi_model_, request.grid_id, 1)
            return bmi_pb2.GetGridPointsResponse(coordinates=self.bmi_model_.get_grid_y(request.grid_id, values))
        except Exception as e:
            self.exception_handler(e, context)

    def getGridZ(self, request, context):
        try:
            values = reserve_grid_nodes(self.bmi_model_, request.grid_id, 2)
            return bmi_pb2.GetGridPointsResponse(coordinates=self.bmi_model_.get_grid_z(request.grid_id, values))
        except Exception as e:
            self.exception_handler(e, context)

    def getGridNodeCount(self, request, context):
        try:
            return bmi_pb2.GetCountResponse(count=self.bmi_model_.get_grid_node_count(request.grid_id))
        except Exception as e:
            self.exception_handler(e, context)

    def getGridEdgeCount(self, request, context):
        try:
            return bmi_pb2.GetCountResponse(count=self.bmi_model_.get_grid_edge_count(request.grid_id))
        except Exception as e:
            self.exception_handler(e, context)

    def getGridFaceCount(self, request, context):
        try:
            return bmi_pb2.GetCountResponse(count=self.bmi_model_.get_grid_face_count(request.grid_id))
        except Exception as e:
            self.exception_handler(e, context)

    def getGridEdgeNodes(self, request, context):
        try:
            size = 2 * self.bmi_model_.get_grid_edge_count(request.grid_id)
            links = numpy.empty(size, dtype=numpy.int64)
            links = self.bmi_model_.get_grid_edge_nodes(request.grid_id, links)
            return bmi_pb2.GetGridEdgeNodesResponse(edge_nodes=links)
        except Exception as e:
            self.exception_handler(e, context)

    def _get_grid_nodes_per_face(self, grid_id):
        size = self.bmi_model_.get_grid_face_count(grid_id)
        links = numpy.empty(size, dtype=numpy.int64)
        return self.bmi_model_.get_grid_nodes_per_face(grid_id, links)

    def getGridFaceNodes(self, request, context):
        try:
            nodes_per_face = self._get_grid_nodes_per_face(request.grid_id)
            size = numpy.sum(nodes_per_face)
            links = numpy.empty(size, dtype=numpy.int64)
            links = self.bmi_model_.get_grid_face_nodes(request.grid_id, links)
            return bmi_pb2.GetGridFaceNodesResponse(face_nodes=links)
        except Exception as e:
            self.exception_handler(e, context)

    def getGridFaceEdges(self, request, context):
        try:
            nodes_per_face = self._get_grid_nodes_per_face(request.grid_id)
            size = numpy.sum(nodes_per_face)
            face_edges = numpy.empty(size, dtype=numpy.int64)
            face_edges = self.bmi_model_.get_grid_face_edges(request.grid_id, face_edges)
            return bmi_pb2.GetGridFaceEdgesResponse(face_edges=face_edges)
        except Exception as e:
            self.exception_handler(e, context)

    def getGridNodesPerFace(self, request, context):
        try:
            nodes_per_face = self._get_grid_nodes_per_face(request.grid_id)
            return bmi_pb2.GetGridNodesPerFaceResponse(nodes_per_face=nodes_per_face)
        except Exception as e:
            self.exception_handler(e, context)

    def __repr__(self):
        # type: (BmiServer) -> str
        return self.bmi_model_.__repr__()



import logging
from unittest.mock import Mock

import grpc
import numpy
import numpy.random
import pytest
from google.protobuf import any_pb2
from google.rpc import error_details_pb2, status_pb2, code_pb2
from grpc_status import rpc_status

from grpc4bmi.bmi_grpc_server import BmiServer
from grpc4bmi.bmi_grpc_client import BmiClient, RemoteException, handle_error
from grpc4bmi.reserve import reserve_values, reserve_grid_shape, reserve_grid_padding
from test.fake_models import SomeException, FailingModel, Rect3DGridModel, UnstructuredGridBmiModel, UniRectGridModel, \
    Rect2DGridModel, Structured3DQuadrilateralsGridModel, Structured2DQuadrilateralsGridModel, Float32Model, Int32Model, \
    BooleanModel
from test.flatbmiheat import FlatBmiHeat

logging.basicConfig(level=logging.DEBUG)

"""
Unit tests for the BMI client class. Every test performs cross-checking with a local instance of the BMI heat toy model.
"""


class MyRpcError(grpc.RpcError):
    def trailing_metadata(self):
        return []


class ServerWrapper(object):

    def __init__(self, server):
        self.server = server
        self.context = Mock(grpc.ServicerContext)

        def forward(status):
            raise MyRpcError(status.details)

        self.context.abort_with_status.side_effect = forward

    def __getattr__(self, item):
        orig_attr = self.server.__getattribute__(item)
        if callable(orig_attr):
            def add_context(*args, **kwargs):
                kwargs["context"] = self.context
                return orig_attr(*args, **kwargs)

            return add_context
        return orig_attr


def make_bmi_classes(init=False):
    client = BmiClient(stub=ServerWrapper(BmiServer(FlatBmiHeat())))
    local = FlatBmiHeat()
    if init:
        numpy.random.seed(0)
        client.initialize(None)
        numpy.random.seed(0)
        local.initialize(None)
    return client, local


def test_server_start():
    client, local = make_bmi_classes()
    assert client is not None
    del client


def test_component_name():
    client, local = make_bmi_classes()
    assert client.get_component_name() == local.get_component_name()
    del client


def test_input_item_count():
    client, local = make_bmi_classes()
    assert client.get_input_item_count() == local.get_input_item_count()
    del client


def test_output_item_count():
    client, local = make_bmi_classes()
    assert client.get_output_item_count() == local.get_output_item_count()
    del client


def test_input_var_names():
    client, local = make_bmi_classes()
    assert client.get_input_var_names() == local.get_input_var_names()
    del client


def test_output_var_names():
    client, local = make_bmi_classes()
    assert client.get_output_var_names() == local.get_output_var_names()
    del client


def test_initialize():
    client, local = make_bmi_classes(True)
    assert client is not None
    client.finalize()
    del client


def test_update():
    client, local = make_bmi_classes(True)
    client.update()
    assert client is not None
    client.finalize()
    del client


def test_update_until():
    client, local = make_bmi_classes(True)
    until = local.get_start_time() + local.get_time_step() + local.get_time_step()
    client.update_until(until)
    assert client.get_current_time() == until
    client.finalize()
    del client


def test_get_time_unit():
    client, local = make_bmi_classes()
    assert client.get_time_units() == local.get_time_units()
    client.finalize()
    del client


def test_get_time_step():
    client, local = make_bmi_classes(True)
    assert client.get_time_step() == local.get_time_step()
    client.finalize()
    del client


def test_get_current_time():
    client, local = make_bmi_classes(True)
    assert client.get_current_time() == local.get_current_time()
    client.finalize()
    del client


def test_get_updated_time():
    client, local = make_bmi_classes(True)
    client.update()
    assert client.get_current_time() != local.get_current_time()
    local.update()
    assert client.get_current_time() == local.get_current_time()
    client.finalize()
    del client


def test_get_start_end_time():
    client, local = make_bmi_classes(True)
    assert client.get_start_time() == local.get_start_time()
    assert client.get_end_time() == local.get_end_time()
    client.finalize()
    del client


def test_get_var_grid():
    client, local = make_bmi_classes(True)
    varname = local.get_output_var_names()[0]
    assert client.get_var_grid(varname) == local.get_var_grid(varname)
    del client


def test_get_var_type():
    client, local = make_bmi_classes(True)
    varname = local.get_output_var_names()[0]
    assert client.get_var_type(varname) == local.get_var_type(varname)
    del client


def test_get_var_units():
    client, local = make_bmi_classes(True)
    varname = local.get_output_var_names()[0]
    assert client.get_var_units(varname) == local.get_var_units(varname)
    del client


def test_get_var_nbytes():
    client, local = make_bmi_classes(True)
    varname = local.get_output_var_names()[0]
    assert client.get_var_nbytes(varname) == local.get_var_nbytes(varname)
    del client


def test_get_var_location():
    client, local = make_bmi_classes(True)
    varname = local.get_output_var_names()[0]

    assert client.get_var_location(varname) == local.get_var_location(varname)

    del client


def test_get_var_value():
    client, local = make_bmi_classes(True)
    varname = local.get_output_var_names()[0]
    actual = client.get_value(varname, reserve_values(client, varname))
    expected = local.get_value(varname, reserve_values(local, varname))
    numpy.testing.assert_allclose(actual, expected)
    del client


def test_get_value_ptr():
    client, local = make_bmi_classes(True)
    varname = local.get_output_var_names()[0]
    with pytest.raises(NotImplementedError):
        client.get_value_ptr(varname)


def test_get_vals_indices():
    client, local = make_bmi_classes(True)
    varname = local.get_output_var_names()[0]
    indices = numpy.array([29, 8, 19, 81])
    result = numpy.empty(len(indices), dtype=client.get_var_type(varname))
    result = client.get_value_at_indices(varname, result, indices)
    expected = numpy.empty(len(indices), dtype=local.get_var_type(varname))
    expected = local.get_value_at_indices(varname, expected, indices)
    numpy.testing.assert_allclose(result, expected)


def test_set_var_value():
    client, local = make_bmi_classes(True)
    varname = local.get_output_var_names()[0]
    values = 0.123 * local.get_value(varname, reserve_values(local, varname))
    client.set_value(varname, values)
    numpy.testing.assert_allclose(client.get_value(varname, reserve_values(client, varname)), values)


def test_set_values_indices():
    client, local = make_bmi_classes(True)
    varname = local.get_output_var_names()[0]
    indices = numpy.array([1, 11, 21])
    values = numpy.array([0.123, 4.567, 8.901], dtype=numpy.float64)

    client.set_value_at_indices(varname, indices, values)

    expected = numpy.empty(len(indices), dtype=client.get_var_type(varname))
    expected = client.get_value_at_indices(varname, expected, indices)
    numpy.testing.assert_allclose(expected, values)


def test_get_grid_size():
    client, local = make_bmi_classes(True)
    varname = local.get_output_var_names()[0]
    grid_id = local.get_var_grid(varname)
    assert client.get_grid_size(grid_id) == local.get_grid_size(grid_id)


def test_get_grid_rank():
    client, local = make_bmi_classes(True)
    varname = local.get_output_var_names()[0]
    grid_id = local.get_var_grid(varname)
    assert client.get_grid_rank(grid_id) == local.get_grid_rank(grid_id)


def test_get_grid_type():
    client, local = make_bmi_classes(True)
    varname = local.get_output_var_names()[0]
    grid_id = local.get_var_grid(varname)
    assert client.get_grid_type(grid_id) == local.get_grid_type(grid_id)


def test_get_grid_shape():
    client, local = make_bmi_classes(True)
    varname = local.get_output_var_names()[0]
    grid_id = local.get_var_grid(varname)

    result = client.get_grid_shape(grid_id, reserve_grid_shape(client, grid_id))

    expected = local.get_grid_shape(grid_id, reserve_grid_shape(local, grid_id))
    numpy.testing.assert_allclose(result, expected)


def test_get_grid_spacing():
    client, local = make_bmi_classes(True)
    varname = local.get_output_var_names()[0]
    grid_id = local.get_var_grid(varname)

    result = client.get_grid_spacing(grid_id, reserve_grid_padding(client, grid_id))

    expected = local.get_grid_spacing(grid_id, reserve_grid_padding(local, grid_id))
    numpy.testing.assert_allclose(result, expected)


def test_get_grid_origin():
    client, local = make_bmi_classes(True)
    varname = local.get_output_var_names()[0]
    grid_id = local.get_var_grid(varname)

    result = client.get_grid_origin(grid_id, reserve_grid_padding(client, grid_id))

    expected = local.get_grid_origin(grid_id, reserve_grid_padding(local, grid_id))
    numpy.testing.assert_allclose(result, expected)


@pytest.mark.parametrize("client_method,client_request", [
    ('initialize', ('config.ini',)),
    ('update', ()),
    ('update_until', (42,)),
    ('finalize', ()),
    ('get_component_name', ()),
    ('get_input_item_count', ()),
    ('get_output_item_count', ()),
    ('get_input_var_names', ()),
    ('get_output_var_names', ()),
    ('get_time_units', ()),
    ('get_time_step', ()),
    ('get_current_time', ()),
    ('get_start_time', ()),
    ('get_end_time', ()),
    ('get_var_grid', ('something',)),
    ('get_var_type', ('something',)),
    ('get_var_itemsize', ('something',)),
    ('get_var_units', ('something',)),
    ('get_var_nbytes', ('something',)),
    ('get_var_location', ('something',)),
    ('get_value', ('something', numpy.empty(0))),
    ('get_value_at_indices', ('something', numpy.empty(0), numpy.array([42]))),
    ('set_value', ('something', numpy.array([1234]))),
    ('set_value_at_indices', ('something', numpy.array([42]), numpy.array([1234]))),
    ('get_grid_size', (42,)),
    ('get_grid_type', (42,)),
    ('get_grid_rank', (42,)),
    ('get_grid_x', (42, numpy.empty(0))),
    ('get_grid_y', (42, numpy.empty(0))),
    ('get_grid_z', (42, numpy.empty(0))),
    ('get_grid_shape', (42, numpy.empty(0))),
    ('get_grid_spacing', (42, numpy.empty(0))),
    ('get_grid_origin', (42, numpy.empty(0))),
    ('get_grid_node_count', (42,)),
    ('get_grid_edge_count', (42,)),
    ('get_grid_face_count', (42,)),
    ('get_grid_edge_nodes', (42, numpy.empty(0))),
    ('get_grid_face_nodes', (42, numpy.empty(0))),
    ('get_grid_face_edges', (42, numpy.empty(0))),
    ('get_grid_nodes_per_face', (42, numpy.empty(0))),
])
def test_method_exception(client_method, client_request):
    exc = SomeException('bmi method always fails')
    model = FailingModel(exc)
    client = BmiClient(stub=ServerWrapper(BmiServer(model)))

    with pytest.raises(Exception) as excinfo:
        getattr(client, client_method)(*client_request)

    assert "bmi method always fails" in str(excinfo.value)


class TestUniRectGridModel:
    @pytest.fixture
    def bmiclient(self):
        model = UniRectGridModel()
        client = BmiClient(stub=ServerWrapper(BmiServer(model)))
        yield client
        del client

    def test_grid_type(self, bmiclient):
        assert bmiclient.get_grid_type(0) == 'uniform_rectilinear'

    def test_grid_size(self, bmiclient):
        assert bmiclient.get_grid_size(0) == 24

    def test_grid_rank(self, bmiclient):
        assert bmiclient.get_grid_rank(0) == 3

    def test_grid_shape(self, bmiclient):
        result = bmiclient.get_grid_shape(0, numpy.empty(3))
        expected = (2, 3, 4)
        numpy.testing.assert_allclose(result, expected)

    def test_grid_origin(self, bmiclient):
        result = bmiclient.get_grid_origin(0, numpy.empty(3))
        expected = (0.1, 1.1, 2.1)
        numpy.testing.assert_allclose(result, expected)

    def test_grid_spacing(self, bmiclient):
        result = bmiclient.get_grid_spacing(0, numpy.empty(3))
        expected = (0.1, 0.2, 0.3)
        numpy.testing.assert_allclose(result, expected)


class TestRect3DGridModel:
    @pytest.fixture
    def bmiclient(self):
        model = Rect3DGridModel()
        client = BmiClient(stub=ServerWrapper(BmiServer(model)))
        yield client
        del client

    def test_grid_size(self, bmiclient):
        assert bmiclient.get_grid_size(0) == 24

    def test_grid_rank(self, bmiclient):
        assert bmiclient.get_grid_rank(0) == 3

    def test_grid_x(self, bmiclient):
        result = bmiclient.get_grid_x(0, numpy.empty(4))
        expected = [0.1, 0.2, 0.3, 0.4]
        numpy.testing.assert_allclose(result, expected)

    def test_grid_y(self, bmiclient):
        result = bmiclient.get_grid_y(0, numpy.empty(3))
        expected = [1.1, 1.2, 1.3]
        numpy.testing.assert_allclose(result, expected)

    def test_grid_z(self, bmiclient):
        result = bmiclient.get_grid_z(0, numpy.empty(2))
        expected = [2.1, 2.2]
        numpy.testing.assert_allclose(result, expected)


class TestRect2DGridModel:
    @pytest.fixture
    def bmiclient(self):
        model = Rect2DGridModel()
        client = BmiClient(stub=ServerWrapper(BmiServer(model)))
        yield client
        del client

    def test_grid_size(self, bmiclient):
        assert bmiclient.get_grid_size(0) == 12

    def test_grid_rank(self, bmiclient):
        assert bmiclient.get_grid_rank(0) == 2

    def test_grid_x(self, bmiclient):
        result = bmiclient.get_grid_x(0, numpy.empty(4))
        expected = [0.1, 0.2, 0.3, 0.4]
        numpy.testing.assert_allclose(result, expected)

    def test_grid_y(self, bmiclient):
        result = bmiclient.get_grid_y(0, numpy.empty(3))
        expected = [1.1, 1.2, 1.3]
        numpy.testing.assert_allclose(result, expected)

    def test_grid_z(self, bmiclient):
        with pytest.raises(MyRpcError) as excinfo:
            bmiclient.get_grid_z(0, numpy.empty(4))

        assert 'out of bounds' in str(excinfo.value)


class TestStructured3DQuadrilateralsGridModel:
    @pytest.fixture
    def bmiclient(self):
        model = Structured3DQuadrilateralsGridModel()
        client = BmiClient(stub=ServerWrapper(BmiServer(model)))
        yield client
        del client

    def test_grid_size(self, bmiclient):
        assert bmiclient.get_grid_size(0) == 4

    def test_grid_rank(self, bmiclient):
        assert bmiclient.get_grid_rank(0) == 3

    def test_grid_shape(self, bmiclient):
        result = bmiclient.get_grid_shape(0, numpy.empty(3))
        expected = [1, 2, 2]
        numpy.testing.assert_allclose(result, expected)

    def test_grid_x(self, bmiclient):
        result = bmiclient.get_grid_x(0, numpy.empty(4))
        expected = [1.1, 0.1, 1.1, 2.1]
        numpy.testing.assert_allclose(result, expected)

    def test_grid_y(self, bmiclient):
        result = bmiclient.get_grid_y(0, numpy.empty(4))
        expected = [2.2, 1.2, 0.2, 2.2]
        numpy.testing.assert_allclose(result, expected)

    def test_grid_z(self, bmiclient):
        result = bmiclient.get_grid_z(0, numpy.empty(4))
        expected = [1.1, 2.2, 3.3, 4.4]
        numpy.testing.assert_allclose(result, expected)


class TestStructured2DQuadrilateralsGridModel:
    @pytest.fixture
    def bmiclient(self):
        model = Structured2DQuadrilateralsGridModel()
        client = BmiClient(stub=ServerWrapper(BmiServer(model)))
        yield client
        del client

    def test_grid_size(self, bmiclient):
        assert bmiclient.get_grid_size(0) == 4

    def test_grid_rank(self, bmiclient):
        assert bmiclient.get_grid_rank(0) == 2

    def test_grid_shape(self, bmiclient):
        result = bmiclient.get_grid_shape(0, numpy.empty(2))
        expected = [2, 2]
        numpy.testing.assert_allclose(result, expected)

    def test_grid_x(self, bmiclient):
        result = bmiclient.get_grid_x(0, numpy.empty(4))
        expected = [1.1, 0.1, 1.1, 2.1]
        numpy.testing.assert_allclose(result, expected)

    def test_grid_y(self, bmiclient):
        result = bmiclient.get_grid_y(0, numpy.empty(4))
        expected = [2.2, 1.2, 0.2, 2.2]
        numpy.testing.assert_allclose(result, expected)

    def test_grid_z(self, bmiclient):
        with pytest.raises(MyRpcError) as excinfo:
            bmiclient.get_grid_z(0, numpy.empty(4))

        assert 'Do not know what z is' in str(excinfo.value)


class TestUnstructuredGridBmiModel:
    @pytest.fixture
    def bmiclient(self):
        model = UnstructuredGridBmiModel()
        client = BmiClient(stub=ServerWrapper(BmiServer(model)))
        yield client
        del client

    def test_get_grid_shape(self, bmiclient):
        with pytest.raises(MyRpcError) as excinfo:
            bmiclient.get_grid_shape(0, numpy.empty(3))

        assert 'Do not know what shape is' in str(excinfo.value)

    def test_grid_size(self, bmiclient):
        assert bmiclient.get_grid_size(0) == 6

    def test_grid_rank(self, bmiclient):
        assert bmiclient.get_grid_rank(0) == 2

    def test_get_grid_node_count(self, bmiclient):
        result = bmiclient.get_grid_node_count(0)

        assert result == 6

    def test_get_grid_edge_count(self, bmiclient):
        result = bmiclient.get_grid_edge_count(0)

        assert result == 8

    def test_get_grid_face_count(self, bmiclient):
        result = bmiclient.get_grid_face_count(0)

        assert result == 3

    def test_get_grid_edge_nodes(self, bmiclient):
        placeholder = numpy.empty(16, dtype=numpy.int)

        result = bmiclient.get_grid_edge_nodes(0, placeholder)

        expected = (0, 1, 1, 2, 2, 3, 3, 0, 1, 4, 4, 5, 5, 2, 5, 3)
        numpy.testing.assert_allclose(result, expected)

    def test_grid_face_nodes(self, bmiclient):
        placeholder = numpy.empty(11, dtype=numpy.int)

        result = bmiclient.get_grid_face_nodes(0, placeholder)

        expected = (0, 1, 2, 3, 1, 4, 5, 2, 2, 5, 3)
        numpy.testing.assert_allclose(result, expected)

    def test_grid_face_edges(self, bmiclient):
        placeholder = numpy.empty(11, dtype=numpy.int)

        result = bmiclient.get_grid_face_edges(0, placeholder)

        expected = (0, 1, 2, 3, 4, 5, 6, 1, 6, 7, 2)
        numpy.testing.assert_allclose(result, expected)

    def test_grid_nodes_per_face(self, bmiclient):
        placeholder = numpy.empty(3, dtype=numpy.int)

        result = bmiclient.get_grid_nodes_per_face(0, placeholder)

        expected = (4, 4, 3)
        numpy.testing.assert_allclose(result, expected)

    def test_grid_x(self, bmiclient):
        result = bmiclient.get_grid_x(0, numpy.empty(6))
        expected = [0., 1., 2., 1., 3., 4.]
        numpy.testing.assert_allclose(result, expected)

    def test_grid_y(self, bmiclient):
        result = bmiclient.get_grid_y(0, numpy.empty(6))
        expected = [3., 1., 2., 4., 0., 3.]
        numpy.testing.assert_allclose(result, expected)

    def test_grid_z(self, bmiclient):
        with pytest.raises(MyRpcError) as excinfo:
            bmiclient.get_grid_z(0, numpy.empty(4))

        assert 'Do not know what z is' in str(excinfo.value)


class TestFloat32Model:
    name = 'plate_surface__temperature'

    @pytest.fixture
    def bmimodel(self):
        return Float32Model()

    @pytest.fixture
    def bmiclient(self, bmimodel):
        client = BmiClient(stub=ServerWrapper(BmiServer(bmimodel)))
        yield client
        del client

    def test_get_value(self, bmiclient):
        result = bmiclient.get_value(self.name, numpy.empty(3))

        expected = numpy.array((1.1, 2.2, 3.3), dtype=numpy.float32)
        numpy.testing.assert_allclose(result, expected)

    def test_get_value_at_indices(self, bmiclient):
        result = bmiclient.get_value_at_indices(self.name, numpy.empty(1), numpy.array([1]))

        expected = numpy.array([2.2], dtype=numpy.float32)
        numpy.testing.assert_allclose(result, expected)

    def test_set_value(self, bmimodel, bmiclient):
        value = numpy.array((2.1, 3.2, 4.3), dtype=numpy.float32)
        bmiclient.set_value(self.name, value)

        numpy.testing.assert_allclose(value, bmimodel.value)

    def test_set_value_at_indices(self, bmimodel, bmiclient):
        value = numpy.array([8.8], dtype=numpy.float32)
        bmiclient.set_value_at_indices(self.name, numpy.array([1]), value)

        expected = numpy.array((1.1, 8.8, 3.3), dtype=numpy.float32)
        numpy.testing.assert_allclose(expected, bmimodel.value)


class TestInt32Model:
    name = 'plate_surface__temperature'

    @pytest.fixture
    def bmimodel(self):
        model = Int32Model()
        yield model
        del model

    @pytest.fixture
    def bmiclient(self, bmimodel):
        client = BmiClient(stub=ServerWrapper(BmiServer(bmimodel)))
        yield client
        del client

    def test_get_value(self, bmiclient):
        result = bmiclient.get_value(self.name, numpy.empty(3))

        expected = numpy.array((12, 24, 36), dtype=numpy.int32)
        numpy.testing.assert_allclose(result, expected)

    def test_get_value_at_indices(self, bmiclient):
        result = bmiclient.get_value_at_indices(self.name, numpy.empty(1), numpy.array([1]))

        expected = numpy.array([24], dtype=numpy.int32)
        numpy.testing.assert_allclose(result, expected)

    def test_set_value(self, bmimodel, bmiclient):
        value = numpy.array((48, 50, 62), dtype=numpy.int32)
        bmiclient.set_value(self.name, value)

        numpy.testing.assert_allclose(value, bmimodel.value)

    def test_set_value_at_indices(self, bmimodel, bmiclient):
        value = numpy.array([88], dtype=numpy.int32)
        bmiclient.set_value_at_indices(self.name, numpy.array([1]), value)

        expected = numpy.array((12, 88, 36), dtype=numpy.int32)
        numpy.testing.assert_allclose(expected, bmimodel.value)


class TestBooleanModel:
    name = 'plate_surface__temperature'

    @pytest.fixture
    def bmimodel(self):
        return BooleanModel()

    @pytest.fixture
    def bmiclient(self, bmimodel):
        client = BmiClient(stub=ServerWrapper(BmiServer(bmimodel)))
        yield client
        del client

    def test_get_value(self, bmiclient):
        with pytest.raises(MyRpcError):
            bmiclient.get_value(self.name, numpy.empty(3))

    def test_get_value_at_indices(self, bmiclient):
        with pytest.raises(MyRpcError):
            bmiclient.get_value_at_indices(self.name, numpy.empty(1, dtype=numpy.bool), numpy.array([1]))

    def test_set_value(self, bmiclient):
        value = numpy.array((False, False, False), dtype=numpy.bool)

        with pytest.raises(NotImplementedError):
            bmiclient.set_value(self.name, value)

    def test_set_value_at_indices(self, bmiclient):
        value = numpy.array([False], dtype=numpy.bool)

        with pytest.raises(NotImplementedError):
            bmiclient.set_value_at_indices(self.name, numpy.array([1]), value)


class MyCall(grpc.RpcError):
    def __init__(self, message, exc, stack_entries):
        super().__init__(message)
        if stack_entries is not None:
            detail = any_pb2.Any()
            detail.Pack(
                error_details_pb2.DebugInfo(
                    stack_entries=stack_entries,
                    detail=repr(exc)
                )
            )
            status = status_pb2.Status(
                code=code_pb2.INTERNAL,
                message=str(exc),
                details=[detail]
            )
        else:
            status = status_pb2.Status(
                code=code_pb2.INTERNAL,
                message=str(exc)
            )
        self.impl = rpc_status.to_status(status)

    def trailing_metadata(self):
        return self.impl.trailing_metadata

    def code(self):
        return self.impl.code

    def details(self):
        return self.impl.details


def test_handle_error_with_stacktrace():
    exc = Exception('Some exception thrown by model on server')
    stack_entries = [
        '  File "/somewhere/on/server/model.py, line 42, in initialize\n  open(fn)',
    ]
    call = MyCall('Server error message', exc, stack_entries)

    with pytest.raises(RemoteException) as excinfo:
        handle_error(call)

        assert excinfo.value.remote_stacktrace == stack_entries
        assert str(excinfo.value) == 'Some exception thrown by model on server'
        assert excinfo.value.__cause == exc


def test_handle_error_without_stacktrace():
    exc = Exception('Some exception thrown by model on server')
    stack_entries = None
    call = MyCall('Server error message', exc, stack_entries)

    with pytest.raises(MyCall) as excinfo:
        try:
            raise call
        except MyCall:
            handle_error(call)

        assert excinfo.value == exc


class OtherCall(grpc.RpcError):
    def __init__(self, message, exc):
        super().__init__(message)
        self.exc = exc

    def trailing_metadata(self):
        return []

    def code(self):
        return code_pb2.INTERNAL

    def details(self):
        return str(self.exc)


def test_handle_error_without_status():
    exc = Exception('Some exception thrown by model on server')
    call = OtherCall('Server error message', exc)

    with pytest.raises(OtherCall) as excinfo:
        try:
            raise call
        except MyCall:
            handle_error(call)

        assert excinfo.value == call


class TestCreateGrpcChannel:
    def test_defaults(self):
        with BmiClient.create_grpc_channel() as channel:
            target = channel._channel.target()
            assert target == b'localhost:50051'

    def test_custom(self):
        with BmiClient.create_grpc_channel(51234, 'somehost') as channel:
            target = channel._channel.target()
            assert target == b'somehost:51234'

    def test_same_port_twice(self):
        port = 51235
        with BmiClient.create_grpc_channel(port) as channel1, BmiClient.create_grpc_channel(port) as channel2:
            assert channel1._channel.target() == b'localhost:51235'
            assert channel2._channel.target() == b'localhost:51235'



import errno
import os
from os.path import abspath
import subprocess
import sys
import logging

from semver import VersionInfo

from grpc4bmi.bmi_grpc_client import BmiClient
from grpc4bmi.utils import stage_config_file

REQUIRED_SINGULARITY_VERSION = '>=3.1.0'


def check_singularity_version():
    p = subprocess.Popen(['singularity', 'version'], stdout=subprocess.PIPE)
    (stdout, _stderr) = p.communicate()
    if p.returncode != 0:
        raise Exception('Unable to determine singularity version')
    if not VersionInfo.parse(stdout.decode('utf-8')).match(REQUIRED_SINGULARITY_VERSION):
        raise Exception(f'Wrong version of singularity found, require version {REQUIRED_SINGULARITY_VERSION}')
    return True


class BmiClientSingularity(BmiClient):
    """BMI GRPC client for singularity server processes
    During initialization launches a singularity container with run-bmi-server as its command.
    The client picks a random port and expects the container to run the server on that port.
    The port is passed to the container using the BMI_PORT environment variable.

    >>> from grpc4bmi.bmi_client_singularity import BmiClientSingularity
    >>> image = 'docker://ewatercycle/wflow-grpc4bmi:latest'
    >>> client = BmiClientSingularity(image, input_dir='wflow_rhine_sbm', output_dir='wflow_output')
    >>> client.initialize('wflow_rhine_sbm/wflow_sbm.ini')
    >>> client.update_until(client.get_end_time())
    >>> del client

    Args:
        image: Singularity image. For Docker Hub image use `docker://*`.
        input_dir (str): Directory for input files of model
        output_dir (str): Directory for input files of model
        timeout (int): Seconds to wait for gRPC client to connect to server
        extra_volumes (Dict[str,str]): Extra volumes to attach to Singularity container.

            The key is the hosts path and the value the mounted volume inside the container.
            Contrary to Docker client, extra volumes are always read/write

            For example:

            .. code-block:: python

                    {'/data/shared/forcings/': '/data/forcings'}

    """
    INPUT_MOUNT_POINT = "/data/input"
    OUTPUT_MOUNT_POINT = "/data/output"

    def __init__(self, image, input_dir=None, output_dir=None, timeout=None, extra_volumes=None):
        check_singularity_version()
        host = 'localhost'
        port = BmiClient.get_unique_port(host)
        args = [
            "singularity",
            "run",
        ]
        mount_points = {} if extra_volumes is None else extra_volumes
        if input_dir is not None:
            mount_points[input_dir] = BmiClientSingularity.INPUT_MOUNT_POINT
            self.input_dir = abspath(input_dir)
        if any(mount_points):
            args += ["--bind", ','.join([hp + ':' + ip for hp, ip in mount_points.items()])]
        if output_dir is not None:
            self.output_dir = abspath(output_dir)
            try:
                # Create output dir ourselves or singularity will complain
                os.mkdir(self.output_dir)
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise e
            args += ["--bind", output_dir + ':' + BmiClientSingularity.OUTPUT_MOUNT_POINT]
        args.append(image)
        env = os.environ.copy()
        env['BMI_PORT'] = str(port)
        logging.info(f'Running {image} singularity container on port {port}')
        self.container = subprocess.Popen(args, env=env, preexec_fn=os.setsid)
        super(BmiClientSingularity, self).__init__(BmiClient.create_grpc_channel(port=port, host=host), timeout=timeout)

    def __del__(self):
        if hasattr(self, "container"):
            self.container.terminate()
            self.container.wait()

    def initialize(self, filename):
        fn = stage_config_file(filename, self.input_dir, self.INPUT_MOUNT_POINT, home_mounted=True)
        super(BmiClientSingularity, self).initialize(fn)

    def get_value_ptr(self, var_name):
        raise NotImplementedError("Cannot exchange memory references across process boundary")



import os
import subprocess
import time

from grpc4bmi.bmi_grpc_client import BmiClient


class BmiClientSubProcess(BmiClient):

    """
    BMI GRPC client that owns its server process, i.e. initiates and destroys the BMI server upon its own construction or
    respective destruction. The server is a forked subprocess running the run_server command.

    >>> from grpc4bmi.bmi_client_subproc import BmiClientSubProcess
    >>> mymodel = BmiClientSubProcess(<PACKAGE>.<MODULE>.<CLASS>)
    """

    def __init__(self, module_name, path=None, timeout=None):
        host = "localhost"
        port = BmiClient.get_unique_port(host)
        name_options = ["--name", module_name]
        port_options = ["--port", str(port)]
        path_options = ["--path", path] if path else []
        self.pipe = subprocess.Popen(["run-bmi-server"] + name_options + port_options + path_options, env=dict(os.environ))
        time.sleep(1)
        super(BmiClientSubProcess, self).__init__(BmiClient.create_grpc_channel(port=port, host=host), timeout=timeout)

    def __del__(self):
        self.pipe.terminate()
        self.pipe.wait()

    def get_value_ref(self, var_name):
        raise NotImplementedError("Cannot exchange memory references across process boundary")



#!/usr/bin/env python

import argparse
import os
import logging
import sys
import time
import signal
import socket
from contextlib import closing
import importlib
from concurrent import futures

import grpc
from grpc_reflection.v1alpha import reflection

from grpc4bmi.bmi_grpc_legacy_server import BmiLegacyServer02
from . import bmi_pb2
from . import bmi_pb2_grpc
from .bmi_grpc_server import BmiServer

try:
    from .bmi_r_model import BmiR
except ImportError:
    BmiR = None

"""
Run server script, turning a BMI implementation into an executable by looping indefinitely, until interrupt signals are
handled. The command line tool needs at least a module and class name to instantiate the BMI wrapper class that exposes
the implementation to other processes.
"""

log = logging.getLogger(__name__)

ENV_BMI_PACKAGE = "BMI_PACKAGE"
ENV_BMI_MODULE = "BMI_MODULE"
ENV_BMI_CLASS = "BMI_CLASS"

kill_server = False


def interrupt(signum, frame):
    global kill_server
    kill_server = True


def build(name, path):
    """Build a model based on it's location and name"""
    if path is not None:
        sys.path.append(path)
    parts = "" if name is None else name.split('.')
    class_name = parts[-1] if len(parts) > 0 else ""
    if not class_name:
        class_name = os.environ.get(ENV_BMI_CLASS, "")
    if not class_name:
        raise ValueError("Missing bmi implementation: class could not be derived from name input %s or environment "
                         "variable %s" % (name, ENV_BMI_CLASS))
    module_name = '.'.join(parts[:-1]) if len(parts) > 1 else ""
    if not module_name:
        module_name = os.environ.get(ENV_BMI_MODULE, "")
    if not module_name:
        raise ValueError("Missing module name: module could not be derived from name input %s or environment "
                         "variable %s" % (name, ENV_BMI_MODULE))

    class_ = getattr(importlib.import_module(module_name), class_name)
    return class_()


def build_r(class_name, source_fn):
    # type: (str, str) -> Bmi
    """"""
    if not BmiR:
        raise ValueError('Missing R dependencies, install with `pip install grpc4bmi[R]')
    return BmiR(class_name, source_fn)


def serve(model, port):
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
    bmi_pb2_grpc.add_BmiServiceServicer_to_server(model, server)
    server.add_insecure_port("[::]:" + str(port))
    service_names = [service.full_name for service in bmi_pb2.DESCRIPTOR.services_by_name.values()]
    service_names.append(reflection.SERVICE_NAME)
    reflection.enable_server_reflection(service_names, server)
    signal.signal(signal.SIGINT, interrupt)
    signal.signal(signal.SIGABRT, interrupt)
    signal.signal(signal.SIGTERM, interrupt)
    log.info("Starting GRPC server for %s at port %d" % (model, port))
    server.start()
    try:
        while not kill_server:
            time.sleep(0.1)
        log.info("Stopping GRPC server for %s at port %d" % (model, port))
        server.stop(0)
    except KeyboardInterrupt:
        log.info("Stopping GRPC server for %s at port %d" % (model, port))
        server.stop(0)


def main(argv=sys.argv[1:]):
    parser = build_parser()

    args = parser.parse_args(argv)

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)

    path = args.path
    if path is None:
        path = os.environ.get("BMI_PATH", None)

    if args.language == "R":
        model = build_r(args.name, path)
    else:
        model = build(args.name, path)

    port = int(os.environ.get("BMI_PORT", 0))
    if port == 0:
        port = args.port
    if port == 0:
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
            s.bind(("", 0))
            port = int(s.getsockname()[1])

    if args.bmi_version == '0.2':
        serve(BmiLegacyServer02(model, args.debug), port)
    else:
        serve(BmiServer(model, args.debug), port)


def build_parser():
    parser = argparse.ArgumentParser(description="BMI GRPC server runner",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--name", "-n", metavar="PACKAGE.MODULE.CLASS", type=str,
                        help="Full name of the BMI implementation class. The module should be in your search "
                             "path and the class should have a constructor with no arguments")
    parser.add_argument("--port", "-p", metavar="N", default=0, type=int,
                        help="Network port for the GRPC server and client. If 0, let the OS choose an available port. "
                             "If the BMI_PORT environment variable is specified, it will take precedence over this "
                             "argument")
    parser.add_argument("--path", "-d", metavar="DIR", default=None, type=str,
                        help="Extra path name to append to the server instance process")
    lang_choices = ['python']
    if BmiR:
        lang_choices.append('R')
    parser.add_argument("--language", default="python", choices=lang_choices,
                        help="Language in which BMI implementation class is written")
    parser.add_argument("--bmi-version", default="1.0.0", choices=["1.0.0", "0.2"],
                        help="Version of BMI interface implemented by model")
    parser.add_argument("--debug", action="store_true",
                        help="Run server in debug mode. "
                             "Logs errors with stacktraces and returns stacktrace in error response")
    return parser


if __name__ == "__main__":
    main()
